import copy
import cPickle

from django.db import models
from django.contrib.auth.models import User, AbstractBaseUser, update_last_login
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.dispatch import receiver
from django.contrib.auth import get_backends


def default_dumpped_dict():
    return cPickle.dumps({})

class SSOUser(AbstractBaseUser):
    username = models.CharField(unique=True, max_length=50)
    extras = models.TextField(default=default_dumpped_dict)

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = []

    def __init__(self, *args, **kwargs):
        fieldsOfModel = [x.name for x in self._meta.fields]  # fields can be saved in database

        _kwargs = copy.deepcopy(kwargs)
        extrainfo = {}
        for fn, fv in filter(lambda item: item[0] not in fieldsOfModel, kwargs.items()):
            _kwargs.pop(fn, None)  # only db fields in _kwargs
            extrainfo[fn] = fv
        _kwargs["extras"] = cPickle.dumps(extrainfo)
        return super(SSOUser, self).__init__(*args, **_kwargs)


    def __getattribute__(self, name):
        val = None
        try:  # read from regular field
            val = super(SSOUser, self).__getattribute__(name)
        except AttributeError as e:  # try to read from extra field
            if name.startswith("__"):  # avoid affecting object stuff
                raise e

            try:
                val = cPickle.loads(str(self.extras)).get(name)
            except Exception as oe:
                pass
        return val


@receiver(user_logged_out, sender=SSOUser)
def notify_backend(request, user, *args, **kwargs):
    from authbackend import SSOAuthBackend
    for b in get_backends():
        if isinstance(b, SSOAuthBackend):
            b.storageengine.remove(user.id)

user_logged_in.disconnect(update_last_login)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SSOUser',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(default=django.utils.timezone.now, verbose_name='last login')),
                ('username', models.CharField(unique=True, max_length=50)),
                ('extras', models.TextField(default=b'{}')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
    ]



import os
from django.conf import settings


def _load_setting(n, default):
    return getattr(settings, n) if hasattr(settings, n) else default

remote_sso_url_prefix = _load_setting("SSO_REMOTE_URL_PREFIX", "/sso/")
REMOTE_REQUEST_TOKEN_URL = os.path.join(remote_sso_url_prefix, "reqeusttoken/")
REMOTE_AUTH_TOKEN_URL = os.path.join(remote_sso_url_prefix, "authtoken/")
REMOTE_SSO_LOGIN_URL = os.path.join(remote_sso_url_prefix, "login/")


# user storage
SSO_USER_STORAGE = _load_setting("SSO_USER_STORAGE", "djssoclient.userstorage.SSOUserDBStorage")
SSO_SETTING_CACHE = _load_setting("SSO_SETTING_CACHE", "default")

__version__ = '0.1'
VERSION = tuple(map(int, __version__.split('.')))



from django.shortcuts import render_to_response
from django.template import RequestContext

def home(request):
    context = RequestContext(request)
    # try:
    #     u = request.user
    #     # print u, u.last_name,
    #     # import cPickle
    #     # ex = cPickle.loads(str(u.extras))
    #     # print ex
    # except:
    #     import traceback
    #     traceback.print_exc()
    #     pass
    return render_to_response("index.html", {"req": request}, context_instance=context)



from django.conf.urls import patterns, include, url
from django.contrib import admin
from django.contrib.auth import views as auth_views
urlpatterns = patterns('',
    url(r'^$', 'ssoclient.views.home', name='home'),

    # login
    url(r'^accounts/login/$', 'djssoclient.views.viewLogin'),
    url(r'^accounts/logout/$', auth_views.logout, {'next_page': '/'}),

    # sso
    url(r'^myssoclient/', include('djssoclient.urls')),
)



import cPickle
from . import SSO_SETTING_CACHE
from django.core.cache import get_cache
from .models import SSOUser


class SSOUserStorage(object):
    class Meta:
        abstract = True

    def save(self, userid, ssouser):
        raise NotImplementedError()

    def find(self, userid):
        raise NotImplementedError()

    def remove(self, userid):
        raise NotImplementedError()


class SSOUserCacheStorage(SSOUserStorage):
    def __init__(self):
        self.cache = get_cache(SSO_SETTING_CACHE)

    def _get_cached_id(self, userid):
        return "sso_user_%s" % userid

    def save(self, userid, ssouser):
        self.cache.set(self._get_cached_id(userid),
                       cPickle.dumps(ssouser))

    def find(self, userid):
        return cPickle.loads(self.cache.get(self._get_cached_id(userid)))

    def remove(self, userid):
        self.cache.delete(self._get_cached_id(userid))


class SSOUserDBStorage(SSOUserStorage):
    def save(self, userid, ssouser):
        ssouser.save()

    def find(self, userid):
        return SSOUser.objects.get(pk=userid)

    def remove(self, userid):
        try:
            SSOUser.objects.get(pk=userid).delete()
        except:
            pass



import os
from setuptools import setup
from djssoclient import __version__

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='dj-sso-client',
    version=__version__,
    packages=['djssoclient'],
    include_package_data=True,
    license='GPL v2.0',
    description='A Django SSO client application, works with dj-sso-server',
    long_description=README,
    url='https://github.com/feifangit/dj-sso-client',
    author='Fan Fei',
    author_email='feifan.pub@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # Replace these appropriately if you are stuck on Python 2.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import urllib
import urllib2
import json
import urlparse
import time
import logging
import hmac
import hashlib
import base64
import abc

from urllib2 import HTTPError
from django.conf import settings

logging.basicConfig()
logger = logging.getLogger("API_CLIENT")
logger.setLevel(logging.DEBUG if settings.DEBUG else logging.INFO)


class AbsAPIClient(object):  # provide overwrite later
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def send_request(self, url, data=None, datafunc=json.loads):
        return


class APIClientHTTPError(Exception):
    def __init__(self, reason, code, *args, **kwargs):
        super(APIClientHTTPError, self).__init__(reason, code, *args, **kwargs)
        self.code = code
        self.message = reason


class APIClient(AbsAPIClient):
    def __init__(self, apikey, seckey, url):
        self.opener = urllib2.build_opener()
        self._baseurl = url
        self._ak = apikey
        self._sk = seckey

    def _sign_msg(self, sk, msg):
        dig = hmac.new(sk, msg, digestmod=hashlib.sha256).digest()
        return base64.b64encode(dig).decode()

    def _sign_url(self, _url):
        url = ("/" + _url) if not _url.startswith("/") else _url
        url_parts = urlparse.urlparse(url)
        qs = urlparse.parse_qs(url_parts.query)
        qs["timestamp"] = time.time()  # UNIX time
        qs["apikey"] = self._ak
        new_qs = urllib.urlencode(qs, True)
        tmpurl = urlparse.urlunparse(list(url_parts[0:4]) + [new_qs] + list(url_parts[5:]))
        final_url = tmpurl + "&signature=" + self._sign_msg(self._sk, tmpurl)  # sign url
        return final_url

    def send_request(self, url, data=None, datafunc=json.loads):
        logger.debug("send to: %s" % urlparse.urljoin(self._baseurl, self._sign_url(url)))
        try:
            resp = self.opener.open(urlparse.urljoin(self._baseurl, self._sign_url(url)), data)
            return resp.code, datafunc(resp.read()) if datafunc else resp.read()
        except HTTPError as e:
            raise APIClientHTTPError(e.code, e.fp.read() if (e.fp and e.code == 403) else e.msg)

client = APIClient(**settings.SSO_API_AUTH_SETTING)
